"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
var index_1 = require("./../../drawboard/Drawboard/index");
var uuid_1 = require("../../utils/uuid");
var types_1 = require("../../markers/types");
var dom_1 = require("../../utils/dom");
require("./index.less");
var prefix = 'fcw-page';
var WhitePage = (function () {
    function WhitePage(source, _a) {
        var _b = _a === void 0 ? {} : _a, mode = _b.mode, whiteboard = _b.whiteboard, parentContainer = _b.parentContainer, extraToolbarItems = _b.extraToolbarItems;
        this.id = uuid_1.uuid();
        this.mode = 'master';
        if (mode) {
            this.mode = mode;
        }
        this.parentContainer = parentContainer;
        this.whiteboard = whiteboard;
        this.initSource(source);
        if (this.mode === 'master') {
            this.initMaster(extraToolbarItems);
        }
        if (this.mode === 'mirror') {
            this.initMirror();
        }
    }
    WhitePage.prototype.open = function () {
        this.drawboard.open();
    };
    WhitePage.prototype.hide = function () {
        this.drawboard.hide();
    };
    WhitePage.prototype.show = function () {
        this.drawboard.show();
    };
    WhitePage.prototype.destroy = function () {
        this.drawboard.destroy();
    };
    WhitePage.prototype.captureSnap = function () {
        var markerSnaps = this.drawboard.markers.map(function (m) { return m.captureSnap(); });
        return {
            id: this.id,
            markers: markerSnaps
        };
    };
    WhitePage.prototype.applySnap = function (snap) {
        var _this = this;
        var markerIdsSet = new Set();
        snap.markers.forEach(function (markerSnap) {
            var marker = _this.drawboard.markerMap[markerSnap.id];
            markerIdsSet.add(markerSnap.id);
            if (marker) {
                marker.applySnap(markerSnap);
            }
            else {
                var newMarker = _this.drawboard.addMarker(types_1.getMarkerByType(markerSnap.type), {
                    id: markerSnap.id
                });
                newMarker.applySnap(markerSnap);
            }
        });
        this.drawboard.markers.forEach(function (marker) {
            if (!markerIdsSet.has(marker.id)) {
                _this.drawboard.deleteMarkerWithEvent(marker);
            }
        });
    };
    WhitePage.prototype.initSource = function (source) {
        if (typeof source.imgSrc === 'string' && !this.parentContainer) {
            throw new Error('Invalid source, If you set image url, you must also set parentContainer');
        }
        this.source = source;
        if (source.imgEle) {
            this.target = source.imgEle;
        }
        if (typeof source.imgSrc === 'string') {
            this.container = dom_1.createDivWithClassName(prefix, this.parentContainer);
            this.container.id = this.id;
            this.target = document.createElement('img');
            this.target.src = source.imgSrc;
            this.target.alt = 'Siema image';
            this.container.appendChild(this.target);
        }
    };
    WhitePage.prototype.initMaster = function (extraToolbarItems) {
        var _this = this;
        if (this.whiteboard) {
            this.drawboard = new index_1.Drawboard({ imgEle: this.target }, {
                extraToolbarItems: extraToolbarItems,
                mountContainer: this.whiteboard.target,
                page: this,
                onChange: function (ev) { return _this.whiteboard.emit(ev); }
            });
        }
        else {
            this.drawboard = new index_1.Drawboard({ imgEle: this.target }, { page: this, mountContainer: this.whiteboard.target });
        }
    };
    WhitePage.prototype.initMirror = function () {
        var _this = this;
        if (!this.whiteboard) {
            throw new Error('Invalid whiteboard');
        }
        this.drawboard = new index_1.Drawboard({ imgEle: this.target }, { page: this, mountContainer: this.whiteboard.target });
        this.whiteboard.eventHub.on('sync', function (ev) {
            try {
                if (ev.target === 'page' && ev.id === _this.id) {
                    _this.onPageSync();
                }
                if (ev.target === 'marker') {
                    _this.onMarkerSync(ev);
                }
            }
            catch (e) {
                console.warn(e);
            }
        });
    };
    WhitePage.prototype.onPageSync = function () { };
    WhitePage.prototype.onMarkerSync = function (ev) {
        if (!ev.marker) {
            return;
        }
        var id = ev.id;
        if (ev.event === 'addMarker' && ev.parentId === this.id) {
            var marker = this.drawboard.markerMap[id];
            if (!marker) {
                this.drawboard.addMarker(types_1.getMarkerByType(ev.marker.type), {
                    id: ev.marker.id,
                    originX: ev.marker.dx,
                    originY: ev.marker.dy
                });
            }
        }
        if (!id) {
            return;
        }
        if (ev.event === 'removeMarker') {
            var marker = this.drawboard.markerMap[id];
            if (marker) {
                this.drawboard.deleteMarker(marker);
            }
        }
        if (ev.event === 'moveMarker' || ev.event === 'resizeMarker') {
            var marker = this.drawboard.markerMap[id];
            if (marker) {
                marker.reactToManipulation(ev.event, ev.marker);
            }
        }
        if (ev.event === 'inputMarker') {
            var marker = this.drawboard.markerMap[id];
            if (marker) {
                marker.setText(ev.marker.text);
            }
        }
    };
    return WhitePage;
}());
exports.WhitePage = WhitePage;
